@@ -19,8 +19,9 @@
 #define LOG_ENABLE_DBG 0
 #include "log.h"
 #include "key-binding.h"
-
-#define ALEN(v) (sizeof(v) / sizeof((v)[0]))
+#include "macros.h"
+#include "xmalloc.h"
+#include "xsnprintf.h"
 
 const struct anchors_map anchors_map[] = {
     {"top-left", ANCHOR_TOP_LEFT},
@@ -44,6 +45,7 @@ static const char *const binding_action_map[] = {
     [BIND_ACTION_CURSOR_LEFT_WORD] = "cursor-left-word",
     [BIND_ACTION_CURSOR_RIGHT] = "cursor-right",
     [BIND_ACTION_CURSOR_RIGHT_WORD] = "cursor-right-word",
+    [BIND_ACTION_DELETE_LINE] = "delete-line",
     [BIND_ACTION_DELETE_PREV] = "delete-prev",
     [BIND_ACTION_DELETE_PREV_WORD] = "delete-prev-word",
     [BIND_ACTION_DELETE_LINE_BACKWARD] = "delete-line-backward",
@@ -51,6 +53,9 @@ static const char *const binding_action_map[] = {
     [BIND_ACTION_DELETE_NEXT_WORD] = "delete-next-word",
     [BIND_ACTION_DELETE_LINE_FORWARD] = "delete-line-forward",
     [BIND_ACTION_INSERT_SELECTED] = "insert-selected",
+    [BIND_ACTION_EXPUNGE] = "expunge",
+    [BIND_ACTION_CLIPBOARD_PASTE] = "clipboard-paste",
+    [BIND_ACTION_PRIMARY_PASTE] = "primary-paste",
     [BIND_ACTION_MATCHES_EXECUTE] = "execute",
     [BIND_ACTION_MATCHES_EXECUTE_OR_NEXT] = "execute-or-next",
     [BIND_ACTION_MATCHES_EXECUTE_INPUT] = "execute-input",
@@ -107,20 +112,19 @@ struct context {
     log_contextual_errno(ctx, __FILE__, __LINE__, __VA_ARGS__)
 
 
-static void __attribute__((format(printf, 5, 6)))
+static void PRINTF(5)
 log_contextual(struct context *ctx, enum log_class log_class,
                const char *file, int lineno, const char *fmt, ...)
 {
     char *formatted_msg = NULL;
     va_list va;
     va_start(va, fmt);
-    if (vasprintf(&formatted_msg, fmt, va) < 0) {
-        va_end(va);
-        return;
-    }
-
+    int ret = vasprintf(&formatted_msg, fmt, va);
     va_end(va);
 
+    if (unlikely(ret < 0))
+        return;
+
     bool print_dot = ctx->key != NULL;
     bool print_colon = ctx->value != NULL;
 
@@ -137,7 +141,7 @@ log_contextual(struct context *ctx, enum log_class log_class,
     free(formatted_msg);
 }
 
-static void __attribute__((format(printf, 4, 5)))
+static void PRINTF(4)
 log_contextual_errno(struct context *ctx, const char *file, int lineno,
                      const char *fmt, ...)
 {
@@ -147,7 +151,7 @@ log_contextual_errno(struct context *ctx, const char *file, int lineno,
     int ret = vasprintf(&formatted_msg, fmt, va);
     va_end(va);
 
-    if (ret < 0)
+    if (unlikely(ret < 0))
         return;
 
     bool print_dot = ctx->key != NULL;
@@ -172,15 +176,6 @@ struct config_file {
     int fd;           /* FD of file, O_RDONLY */
 };
 
-static const char *
-get_user_home_dir(void)
-{
-    const struct passwd *passwd = getpwuid(getuid());
-    if (passwd == NULL)
-        return NULL;
-    return passwd->pw_dir;
-}
-
 static struct config_file
 open_config(void)
 {
@@ -189,20 +184,14 @@ open_config(void)
 
     const char *xdg_config_home = getenv("XDG_CONFIG_HOME");
     const char *xdg_config_dirs = getenv("XDG_CONFIG_DIRS");
-    const char *home_dir = get_user_home_dir();
+    const char *home_dir = getenv("HOME");
     char *xdg_config_dirs_copy = NULL;
 
     /* First, check XDG_CONFIG_HOME (or .config, if unset) */
-    if (xdg_config_home != NULL && xdg_config_home[0] != '\0') {
-        if (asprintf(&path, "%s/fuzzel/fuzzel.ini", xdg_config_home) < 0) {
-            LOG_ERRNO("failed to build fuzzel.ini path");
-            goto done;
-        }
+    if (xdg_config_home != NULL && xdg_config_home[0] == '/') {
+        path = xstrjoin(xdg_config_home, "/fuzzel/fuzzel.ini");
     } else if (home_dir != NULL) {
-        if (asprintf(&path, "%s/.config/fuzzel/fuzzel.ini", home_dir) < 0) {
-            LOG_ERRNO("failed to build fuzzel.ini path");
-            goto done;
-        }
+        path = xstrjoin(home_dir, "/.config/fuzzel/fuzzel.ini");
     }
 
     if (path != NULL) {
@@ -216,8 +205,8 @@ open_config(void)
     }
 
     xdg_config_dirs_copy = xdg_config_dirs != NULL && xdg_config_dirs[0] != '\0'
-        ? strdup(xdg_config_dirs)
-        : strdup("/etc/xdg");
+        ? xstrdup(xdg_config_dirs)
+        : xstrdup("/etc/xdg");
 
     if (xdg_config_dirs_copy == NULL || xdg_config_dirs_copy[0] == '\0')
         goto done;
@@ -227,13 +216,7 @@ open_config(void)
          conf_dir = strtok(NULL, ":"))
     {
         free(path);
-        path = NULL;
-
-        if (asprintf(&path, "%s/fuzzel/fuzzel.ini", conf_dir) < 0) {
-            LOG_ERRNO("failed to build fuzzel.ini path");
-            goto done;
-        }
-
+        path = xstrjoin(conf_dir, "/fuzzel/fuzzel.ini");
         int fd = open(path, O_RDONLY | O_CLOEXEC);
         if (fd >= 0) {
             ret = (struct config_file){.path = path, .fd = fd};
@@ -268,7 +251,7 @@ parse_modifiers(struct context *ctx, const char *text, size_t len,
     if (strncmp(text, "none", len) == 0)
         return true;
 
-    char *copy = strndup(text, len);
+    char *copy = xstrndup(text, len);
 
     for (char *tok_ctx = NULL, *key = strtok_r(copy, "+", &tok_ctx);
          key != NULL;
@@ -347,7 +330,7 @@ value_to_key_combos(struct context *ctx, int action,
 
     struct config_key_binding new_combos[combo_count];
 
-    char *copy = strdup(ctx->value);
+    char *copy = xstrdup(ctx->value);
     size_t idx = 0;
 
     for (char *tok_ctx = NULL, *combo = strtok_r(copy, " ", &tok_ctx);
@@ -389,9 +372,9 @@ value_to_key_combos(struct context *ctx, int action,
 
     remove_from_key_bindings_list(bindings, action);
 
-    bindings->arr = realloc(
+    bindings->arr = xreallocarray(
         bindings->arr,
-        (bindings->count + combo_count) * sizeof(bindings->arr[0]));
+        bindings->count + combo_count, sizeof(bindings->arr[0]));
 
     memcpy(&bindings->arr[bindings->count],
            new_combos,
@@ -420,15 +403,13 @@ modifiers_equal(const struct config_key_modifiers *mods1,
 static char *
 modifiers_to_str(const struct config_key_modifiers *mods)
 {
-    char *ret = NULL;
-    if (asprintf(&ret,
-                 "%s%s%s%s",
-                 mods->ctrl ? XKB_MOD_NAME_CTRL "+" : "",
-                 mods->alt ? XKB_MOD_NAME_ALT "+": "",
-                 mods->super ? XKB_MOD_NAME_LOGO "+": "",
-                 mods->shift ? XKB_MOD_NAME_SHIFT "+": "") < 0)
-        return NULL;
-    return ret;
+    return xasprintf(
+        "%s%s%s%s",
+        mods->ctrl ? XKB_MOD_NAME_CTRL "+" : "",
+        mods->alt ? XKB_MOD_NAME_ALT "+": "",
+        mods->super ? XKB_MOD_NAME_LOGO "+": "",
+        mods->shift ? XKB_MOD_NAME_SHIFT "+": ""
+    );
 }
 
 static bool
@@ -539,11 +520,24 @@ str_to_uint32(const char *s, int base, uint32_t *res)
     return true;
 }
 
+static bool
+str_to_uint16(const char *s, int base, uint16_t *res)
+{
+    unsigned long v;
+    bool ret = str_to_ulong(s, base, &v);
+    if (!ret)
+        return false;
+    if (v > UINT16_MAX)
+        return false;
+    *res = v;
+    return true;
+}
+
 static bool
 value_to_str(struct context *ctx, char **res)
 {
     free(*res);
-    *res = strdup(ctx->value);
+    *res = xstrdup(ctx->value);
     return true;
 }
 
@@ -596,6 +590,17 @@ value_to_uint32(struct context *ctx, int base, uint32_t *res)
     return true;
 }
 
+static bool
+value_to_uint16(struct context *ctx, int base, uint16_t *res)
+{
+    if (!str_to_uint16(ctx->value, base, res)){
+        LOG_CONTEXTUAL_ERR(
+            "invalid integer value, or outside range 0-%u", UINT32_MAX);
+        return false;
+    }
+    return true;
+}
+
 static bool
 value_to_color(struct context *ctx, bool allow_alpha, struct rgba *color)
 {
@@ -691,7 +696,7 @@ value_to_enum(struct context *ctx, const char **value_map, int *res)
     assert(size < sizeof(valid_values));
 
     for (size_t i = 0; i < count; i++)
-        idx += snprintf(&valid_values[idx], size - idx, "'%s', ", value_map[i]);
+        idx += xsnprintf(&valid_values[idx], size - idx, "'%s', ", value_map[i]);
 
     if (count > 0)
         valid_values[idx - 2] = '\0';
@@ -725,10 +730,7 @@ parse_section_main(struct context *ctx)
                 return false;
             }
 
-            if (asprintf(&_include_path, "%s/%s", home_dir, value + 2) < 0) {
-                LOG_ERRNO("failed to allocate path for include directive");
-                return false;
-            }
+            _include_path = xstrjoin3(home_dir, "/", value + 2);
             include_path = _include_path;
         } else
             include_path = value;
@@ -739,7 +741,7 @@ parse_section_main(struct context *ctx)
             return false;
         }
 
-        FILE *include = fopen(include_path, "r");
+        FILE *include = fopen(include_path, "re");
 
         if (include == NULL) {
             LOG_CONTEXTUAL_ERRNO("failed to open");
@@ -756,12 +758,18 @@ parse_section_main(struct context *ctx)
         return ret;
     }
 
+    else if (strcmp(key, "namespace") == 0)
+        return value_to_str(ctx, &conf->namespace);
+
     else if (strcmp(key, "output") == 0)
         return value_to_str(ctx, &conf->output);
 
     else if (strcmp(key, "font") == 0)
         return value_to_str(ctx, &conf->font);
 
+    else if (strcmp(key, "use-bold") == 0)
+        return value_to_bool(ctx, &conf->use_bold);
+
     else if (strcmp(key, "dpi-aware") == 0) {
         if (strcmp(value, "auto") == 0)
             conf->dpi_aware = DPI_AWARE_AUTO;
@@ -774,15 +782,30 @@ parse_section_main(struct context *ctx)
         return true;
     }
 
+    else if (strcmp(key, "gamma-correct-blending") == 0)
+        return value_to_bool(ctx, &conf->gamma_correct);
+
+    else if (strcmp(key, "render-workers") == 0)
+        return value_to_uint16(ctx, 10, &conf->render_worker_count);
+
+    else if (strcmp(key, "match-workers") == 0)
+        return value_to_uint16(ctx, 10, &conf->match_worker_count);
+
     else if (strcmp(key, "prompt") == 0)
         return value_to_wchars(ctx, &conf->prompt);
 
+    else if (strcmp(key, "placeholder") == 0)
+        return value_to_wchars(ctx, &conf->placeholder);
+
     else if (strcmp(key, "icon-theme") == 0)
         return value_to_str(ctx, &conf->icon_theme);
 
     else if (strcmp(key, "icons-enabled") == 0)
         return value_to_bool(ctx, &conf->icons_enabled);
 
+    else if (strcmp(key, "hide-before-typing") == 0)
+        return value_to_bool(ctx, &conf->hide_when_prompt_empty);
+
     else if (strcmp(key, "list-executables-in-path") == 0)
         return value_to_bool(ctx, &conf->list_executables_in_path);
 
@@ -792,7 +815,7 @@ parse_section_main(struct context *ctx)
 
         enum match_fields match_fields = 0;
 
-        char *copy = strdup(value);
+        char *copy = xstrdup(value);
         for (const char *field = strtok(copy, ",");
              field != NULL;
              field = strtok(NULL, ","))
@@ -812,7 +835,7 @@ parse_section_main(struct context *ctx)
 
             enum match_fields field_value = 0;
 
-            for (size_t i = 0; i < sizeof(map) / sizeof(map[0]); i++) {
+            for (size_t i = 0; i < ALEN(map); i++) {
                 if (strcmp(field, map[i].name) == 0) {
                     field_value = map[i].value;
                     break;
@@ -861,8 +884,27 @@ parse_section_main(struct context *ctx)
     else if (strcmp(key, "filter-desktop") == 0)
         return value_to_bool(ctx, &conf->filter_desktop);
 
-    else if (strcmp(key, "fuzzy") == 0)
-        return value_to_bool(ctx, &conf->fuzzy.enabled);
+    else if (strcmp(key, "match-mode") == 0) {
+        _Static_assert(sizeof(conf->match_mode) == sizeof(int),
+                       "enum is not 32-bit");
+
+        return value_to_enum(
+            ctx,
+            (const char *[]){"exact", "fzf", "fuzzy", NULL},
+            (int *)&conf->match_mode);
+    }
+
+    else if (strcmp(key, "sort-result") == 0)
+        return value_to_bool(ctx, &conf->sort_result);
+
+    else if (strcmp(key, "match-counter") == 0)
+        return value_to_bool(ctx, &conf->match_counter);
+
+    else if (strcmp(key, "delayed-filter-ms") == 0)
+        return value_to_uint32(ctx, 10, &conf->delayed_filter_ms);
+
+    else if (strcmp(key, "delayed-filter-limit") == 0)
+        return value_to_uint32(ctx, 10, &conf->delayed_filter_limit);
 
     else if (strcmp(key, "show-actions") == 0)
         return value_to_bool(ctx, &conf->actions_enabled);
@@ -899,10 +941,21 @@ parse_section_main(struct context *ctx)
         conf->anchor = anchor;
         return true;
     }
+    else if (strcmp(key, "x-margin") == 0)
+      return value_to_uint32(ctx, 10, &conf->margin.x);
+
+    else if (strcmp(key, "y-margin") == 0)
+      return value_to_uint32(ctx, 10, &conf->margin.y);
 
     else if (strcmp(key, "lines") == 0)
         return value_to_uint32(ctx, 10, &conf->lines);
 
+    else if (strcmp(key, "minimal-lines") == 0)
+        return value_to_bool(ctx, &conf->minimal_lines);
+
+    else if (strcmp(key, "hide-prompt") == 0)
+        return value_to_bool(ctx, &conf->hide_prompt);
+
     else if (strcmp(key, "width") == 0)
         return value_to_uint32(ctx, 10, &conf->chars);
 
@@ -938,6 +991,18 @@ parse_section_main(struct context *ctx)
         return true;
     }
 
+    else if (strcmp(key, "scaling-filter") == 0) {
+        _Static_assert(sizeof(conf->png_scaling_filter) == sizeof(int),
+            "enum is not 32-bit");
+
+        return value_to_enum(
+            ctx,
+            (const char *[]){"none", "nearest", "bilinear", "box", "linear",
+                             "cubic", "lanczos2", "lanczos3",
+                             "lanczos3-stretched", NULL},
+            (int *)&conf->png_scaling_filter);
+    }
+
     else if (strcmp(key, "layer") == 0) {
         if (strcasecmp(value, "top") == 0) {
             conf->layer = ZWLR_LAYER_SHELL_V1_LAYER_TOP;
@@ -947,13 +1012,34 @@ parse_section_main(struct context *ctx)
             return true;
         }
 
-        LOG_CONTEXTUAL_ERR("not one of 'top', 'overay'");
+        LOG_CONTEXTUAL_ERR("not one of 'top', 'overlay'");
+        return false;
+    }
+
+    else if (strcmp(key, "keyboard-focus") == 0) {
+        if (strcasecmp(value, "exclusive") == 0) {
+            conf->keyboard_focus = ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_EXCLUSIVE;
+            return true;
+        } else if (strcasecmp(value, "on-demand") == 0) {
+            conf->keyboard_focus = ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_ON_DEMAND;
+            return true;
+        }
+        LOG_CONTEXTUAL_ERR("not one of 'exclusive', 'on-demand'");
         return false;
     }
 
     else if (strcmp(key, "exit-on-keyboard-focus-loss") == 0)
         return value_to_bool(ctx, &conf->exit_on_kb_focus_loss);
 
+    else if (strcmp(key, "cache") == 0)
+        return value_to_str(ctx, &conf->cache_path);
+
+    else if (strcmp(key, "auto-select") == 0)
+        return value_to_bool(ctx, &conf->auto_select);
+
+    else if (strcmp(key, "enable-mouse") == 0)
+        return value_to_bool(ctx, &conf->enable_mouse);
+
     else
         LOG_CONTEXTUAL_ERR("not a valid option: %s", key);
 
@@ -972,6 +1058,15 @@ parse_section_colors(struct context *ctx)
     else if (strcmp(key, "text") == 0)
         return value_to_color(ctx, true, &conf->colors.text);
 
+    else if (strcmp(key, "prompt") == 0)
+        return value_to_color(ctx, true, &conf->colors.prompt);
+
+    else if (strcmp(key, "placeholder") == 0)
+        return value_to_color(ctx, true, &conf->colors.placeholder);
+
+    else if (strcmp(key, "input") == 0)
+        return value_to_color(ctx, true, &conf->colors.input);
+
     else if (strcmp(key, "match") == 0)
         return value_to_color(ctx, true, &conf->colors.match);
 
@@ -984,6 +1079,9 @@ parse_section_colors(struct context *ctx)
     else if (strcmp(key, "selection-match") == 0)
         return value_to_color(ctx, true, &conf->colors.selection_match);
 
+    else if (strcmp(key, "counter") == 0)
+        return value_to_color(ctx, true, &conf->colors.counter);
+
     else if (strcmp(key, "border") == 0)
         return value_to_color(ctx, true, &conf->colors.border);
 
@@ -1003,15 +1101,10 @@ parse_section_border(struct context *ctx)
         return value_to_uint32(ctx, 10, &conf->border.size);
 
     else if (strcmp(key, "radius") == 0)
-#if defined(FUZZEL_ENABLE_CAIRO)
         return value_to_uint32(ctx, 10, &conf->border.radius);
-#else
-    {
-        LOG_CONTEXTUAL_WARN(
-            "fuzzel compiled without cairo support; ignoring");
-        return true;
-    }
-#endif
+
+    else if (strcmp(key, "selection-radius") == 0)
+        return value_to_uint32(ctx, 10, &conf->selection_border.radius);
 
     else
         LOG_CONTEXTUAL_ERR("not a valid option: %s", key);
@@ -1217,7 +1310,7 @@ parse_config_file(FILE *f, struct config *conf, const char *path,
             continue;                           \
     }
 
-    char *section_name = strdup("main");
+    char *section_name = xstrdup("main");
 
     struct context context = {
         .conf = conf,
@@ -1311,7 +1404,7 @@ parse_config_file(FILE *f, struct config *conf, const char *path,
             }
 
             free(section_name);
-            section_name = strdup(key_value);
+            section_name = xstrdup(key_value);
             context.section = section_name;
 
             /* Process next line */
@@ -1423,6 +1516,8 @@ add_default_key_bindings(struct config *conf)
     static const struct config_key_binding bindings[] = {
         {BIND_ACTION_CANCEL, m_none, {{XKB_KEY_Escape}}},
         {BIND_ACTION_CANCEL, m_ctrl, {{XKB_KEY_g}}},
+        {BIND_ACTION_CANCEL, m_ctrl, {{XKB_KEY_c}}},
+        {BIND_ACTION_CANCEL, m_ctrl, {{XKB_KEY_bracketleft}}},
 
         {BIND_ACTION_CURSOR_HOME, m_none, {{XKB_KEY_Home}}},
         {BIND_ACTION_CURSOR_HOME, m_ctrl, {{XKB_KEY_a}}},
@@ -1442,9 +1537,13 @@ add_default_key_bindings(struct config *conf)
         {BIND_ACTION_CURSOR_RIGHT_WORD, m_alt, {{XKB_KEY_f}}},
         {BIND_ACTION_CURSOR_RIGHT_WORD, m_ctrl, {{XKB_KEY_Right}}},
 
+        {BIND_ACTION_DELETE_LINE, m_ctrl_shift, {{XKB_KEY_BackSpace}}},
+
         {BIND_ACTION_DELETE_PREV, m_none, {{XKB_KEY_BackSpace}}},
+        {BIND_ACTION_DELETE_PREV, m_ctrl, {{XKB_KEY_h}}},
 
         {BIND_ACTION_DELETE_PREV_WORD, m_ctrl, {{XKB_KEY_BackSpace}}},
+        {BIND_ACTION_DELETE_PREV_WORD, m_ctrl, {{XKB_KEY_w}}},
         {BIND_ACTION_DELETE_PREV_WORD, m_alt, {{XKB_KEY_BackSpace}}},
 
         {BIND_ACTION_DELETE_NEXT, m_none, {{XKB_KEY_Delete}}},
@@ -1460,6 +1559,14 @@ add_default_key_bindings(struct config *conf)
 
         {BIND_ACTION_INSERT_SELECTED, m_ctrl, {{XKB_KEY_Tab}}},
 
+        {BIND_ACTION_EXPUNGE, m_shift, {{XKB_KEY_Delete}}},
+        {BIND_ACTION_EXPUNGE, m_shift, {{XKB_KEY_KP_Delete}}},
+
+        {BIND_ACTION_CLIPBOARD_PASTE, m_ctrl, {{XKB_KEY_v}}},
+        {BIND_ACTION_CLIPBOARD_PASTE, m_none, {{XKB_KEY_XF86Paste}}},
+        {BIND_ACTION_PRIMARY_PASTE, m_shift, {{XKB_KEY_Insert}}},
+        {BIND_ACTION_PRIMARY_PASTE, m_shift, {{XKB_KEY_KP_Insert}}},
+
         {BIND_ACTION_MATCHES_EXECUTE, m_none, {{XKB_KEY_Return}}},
         {BIND_ACTION_MATCHES_EXECUTE, m_none, {{XKB_KEY_KP_Enter}}},
         {BIND_ACTION_MATCHES_EXECUTE, m_ctrl, {{XKB_KEY_y}}},
@@ -1507,8 +1614,7 @@ add_default_key_bindings(struct config *conf)
     };
 
     conf->key_bindings.count = ALEN(bindings);
-    conf->key_bindings.arr = malloc(sizeof(bindings));
-    memcpy(conf->key_bindings.arr, bindings, sizeof(bindings));
+    conf->key_bindings.arr = xmemdup(bindings, sizeof(bindings));
 }
 
 bool
@@ -1518,7 +1624,10 @@ config_load(struct config *conf, const char *conf_path,
     bool ret = !errors_are_fatal;
     *conf = (struct config) {
         .output = NULL,
-        .prompt = c32dup(U"> "),
+        .namespace = xstrdup("launcher"),
+        .prompt = xc32dup(U"> "),
+        .placeholder = xc32dup(U""),
+        .search_text = xc32dup(U""),
         .match_fields = MATCH_FILENAME | MATCH_NAME | MATCH_GENERIC,
         .password_mode = {
             .character = U'\0',
@@ -1526,24 +1635,41 @@ config_load(struct config *conf, const char *conf_path,
         },
         .terminal = NULL,
         .launch_prefix = NULL,
-        .font = strdup("monospace"),
+        .font = xstrdup("monospace"),
+        .use_bold = false,
         .dpi_aware = DPI_AWARE_AUTO,
+        .gamma_correct = false,
+        .render_worker_count = sysconf(_SC_NPROCESSORS_ONLN),
+        .match_worker_count = sysconf(_SC_NPROCESSORS_ONLN),
+        .filter_desktop = false,
         .icons_enabled = true,
-        .icon_theme = strdup("hicolor"),
+        .icon_theme = xstrdup("default"),
+        .hide_when_prompt_empty = false,
         .actions_enabled = false,
+        .match_mode = MATCH_MODE_FZF,
+        .sort_result = true,
+        .match_counter = false,
+        .delayed_filter_ms = 300,
+        .delayed_filter_limit = 20000,
         .fuzzy = {
             .min_length = 3,
             .max_length_discrepancy = 2,
             .max_distance = 1,
-            .enabled = true,
         },
         .dmenu = {
             .enabled = false,
             .mode = DMENU_MODE_TEXT,
             .exit_immediately_if_empty = false,
             .delim = '\n',
+            .nth_delim = '\t',
+            .with_nth_format = NULL,
+            .accept_nth_format = NULL,
         },
         .anchor = ANCHOR_CENTER,
+        .margin = {
+            .x = 0,
+            .y = 0,
+        },
         .lines = 15,
         .chars = 30,
         .tabs = 8,
@@ -1553,24 +1679,38 @@ config_load(struct config *conf, const char *conf_path,
             .inner = 0,
         },
         .colors = {
-            .background = conf_hex_to_rgba(0xfdf6e3dd),
+            .background = conf_hex_to_rgba(0xfdf6e3ff),
             .border = conf_hex_to_rgba(0x002b36ff),
             .text = conf_hex_to_rgba(0x657b83ff),
+            .prompt = conf_hex_to_rgba(0x586e75ff),
+            .input = conf_hex_to_rgba(0x657b83ff),
             .match = conf_hex_to_rgba(0xcb4b16ff),
-            .selection = conf_hex_to_rgba(0xeee8d5dd),
-            .selection_text = conf_hex_to_rgba(0x657b83ff),
+            .selection = conf_hex_to_rgba(0xeee8d5ff),
+            .selection_text = conf_hex_to_rgba(0x586e75ff),
             .selection_match = conf_hex_to_rgba(0xcb4b16ff),
+            .counter = conf_hex_to_rgba(0x93a1a1ff),
+            .placeholder = conf_hex_to_rgba(0x93a1a1ff),
         },
         .border = {
             .size = 1u,
             .radius = 10u,
         },
+        .selection_border = {
+            .size = 0u,
+            .radius = 0u,
+        },
         .image_size_ratio = 0.5,
+        .png_scaling_filter = SCALING_FILTER_BOX,
         .line_height = {-1, 0.0},
         .letter_spacing = {0},
-        .layer = ZWLR_LAYER_SHELL_V1_LAYER_TOP,
+        .layer = ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY,
+        .keyboard_focus = ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_EXCLUSIVE,
         .exit_on_kb_focus_loss = true,
         .list_executables_in_path = false,
+        .cache_path = NULL,
+        .auto_select = false,
+        .print_timing_info = false,
+        .enable_mouse = true,
     };
 
     add_default_key_bindings(conf);
@@ -1579,20 +1719,17 @@ config_load(struct config *conf, const char *conf_path,
 
     const char *terminal = getenv("TERMINAL");
     if (terminal != NULL) {
-        if (asprintf(&conf->terminal, "%s -e", terminal) < 0) {
-            LOG_ERRNO("failed to build default terminal command line");
-            goto out;
-        }
+        conf->terminal = xstrjoin(terminal, " -e");
     }
 
     if (conf_path != NULL) {
-        int fd = open(conf_path, O_RDONLY);
+        int fd = open(conf_path, O_RDONLY | O_CLOEXEC);
         if (fd < 0) {
             LOG_ERRNO("%s: failed to open", conf_path);
             goto out;
         }
 
-        conf_file.path = strdup(conf_path);
+        conf_file.path = xstrdup(conf_path);
         conf_file.fd = fd;
     } else {
         conf_file = open_config();
@@ -1619,6 +1756,7 @@ config_load(struct config *conf, const char *conf_path,
         ret = true;
 
     fclose(f);
+    conf_file.fd = -1;
 
 out:
 
@@ -1632,12 +1770,18 @@ out:
 void
 config_free(struct config *conf)
 {
+    free(conf->namespace);
     free(conf->output);
     free(conf->prompt);
+    free(conf->placeholder);
+    free(conf->search_text);
     free(conf->terminal);
     free(conf->launch_prefix);
     free(conf->font);
     free(conf->icon_theme);
+    free(conf->cache_path);
+    free(conf->dmenu.with_nth_format);
+    free(conf->dmenu.accept_nth_format);
     free_key_binding_list(&conf->key_bindings);
 }
 
