@@ -1,5 +1,6 @@
 #include "xdg.h"
 
+#include <assert.h>
 #include <ctype.h>
 #include <errno.h>
 #include <limits.h>
@@ -21,8 +22,11 @@
 #include "log.h"
 #include "char32.h"
 #include "icon.h"
+#include "macros.h"
+#include "xmalloc.h"
+#include "xsnprintf.h"
 
-typedef tll(struct application) application_llist_t;
+typedef tll(struct application *) application_llist_t;
 
 struct locale_variants {
     char *lang_country_modifier;
@@ -54,6 +58,7 @@ struct action {
     char *path;
     bool visible;
     bool use_terminal;
+    bool no_startup_notify;
 };
 
 static bool
@@ -80,15 +85,17 @@ filter_desktop_entry(const struct action *act, const char_list_t *desktops)
 }
 
 static void
-parse_desktop_file(int fd, char *id, const char32_t *file_basename,
+parse_desktop_file(int fd, char *id, const char32_t *file_basename_lowercase,
                    const char *terminal, bool include_actions,
                    bool filter_desktops, char_list_t *desktops,
                    const struct locale_variants *lc_messages,
                    application_llist_t *applications)
 {
-    FILE *f = fdopen(fd, "r");
-    if (f == NULL)
+    FILE *f = fdopen(fd, "re");
+    if (f == NULL) {
+        close(fd);
         return;
+    }
 
     bool is_desktop_entry = false;
 
@@ -161,20 +168,20 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename,
 
                         action->generic_name =
                             default_action->generic_name != NULL
-                            ? c32dup(default_action->generic_name) : NULL;
+                            ? xc32dup(default_action->generic_name) : NULL;
                         action->comment = default_action->comment != NULL
-                            ? c32dup(default_action->comment) : NULL;
+                            ? xc32dup(default_action->comment) : NULL;
                         action->path = default_action->path != NULL
-                            ? strdup(default_action->path) : NULL;
+                            ? xstrdup(default_action->path) : NULL;
                         action->icon = default_action->icon != NULL
-                            ? strdup(default_action->icon) : NULL;
+                            ? xstrdup(default_action->icon) : NULL;
                         action->visible = default_action->visible;
                         action->use_terminal = default_action->use_terminal;
 
                         tll_foreach(default_action->keywords, it)
-                            tll_push_back(action->keywords, c32dup(it->item));
+                            tll_push_back(action->keywords, xc32dup(it->item));
                         tll_foreach(default_action->categories, it)
-                            tll_push_back(action->categories, c32dup(it->item));
+                            tll_push_back(action->categories, xc32dup(it->item));
 
                         action_is_valid = true;
                         break;
@@ -257,13 +264,13 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename,
             else if (strcmp(key, "Exec") == 0) {
                 free(action->exec);
                 free(action->wexec);
-                action->exec = strdup(value);
+                action->exec = xstrdup(value);
                 action->wexec = ambstoc32(value);
             }
 
             else if (strcmp(key, "Path") == 0) {
                 free(action->path);
-                action->path = strdup(value);
+                action->path = xstrdup(value);
             }
 
             else if (strcmp(key, "GenericName") == 0) {
@@ -274,9 +281,14 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename,
                 }
             }
 
+            else if (strcmp(key, "StartupNotify") == 0) {
+                if (strcmp(value, "false"))
+                    action->no_startup_notify = true;
+            }
+
             else if (strcmp(key, "StartupWMClass") == 0) {
                 free(action->app_id);
-                action->app_id = strdup(value);
+                action->app_id = xstrdup(value);
             }
 
             else if (strcmp(key, "Comment") == 0) {
@@ -322,7 +334,7 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename,
                      action != NULL;
                      action = strtok_r(NULL, ";", &ctx))
                 {
-                    tll_push_back(action_names, strdup(action));
+                    tll_push_back(action_names, xstrdup(action));
                 }
             }
 
@@ -331,7 +343,7 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename,
                      desktop != NULL;
                      desktop = strtok_r(NULL, ";", &ctx))
                 {
-                    tll_push_back(action->onlyshowin, strdup(desktop));
+                    tll_push_back(action->onlyshowin, xstrdup(desktop));
                 }
             }
 
@@ -340,13 +352,13 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename,
                      desktop != NULL;
                      desktop = strtok_r(NULL, ";", &ctx))
                 {
-                    tll_push_back(action->notshowin, strdup(desktop));
+                    tll_push_back(action->notshowin, xstrdup(desktop));
                 }
             }
 
             else if (strcmp(key, "Icon") == 0) {
                 free(action->icon);
-                action->icon = strdup(value);
+                action->icon = xstrdup(value);
             }
 
             else if (strcmp(key, "Hidden") == 0 ||
@@ -391,46 +403,124 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename,
             a->name = ambstoc32("<no title>");
 
         if (a->use_terminal && terminal != NULL && a->exec != NULL) {
-            char *exec_with_terminal = malloc(
-                strlen(terminal) + 1 + strlen(a->exec) + 1);
-            strcpy(exec_with_terminal, terminal);
-            strcat(exec_with_terminal, " ");
-            strcat(exec_with_terminal, a->exec);
+            char *exec_with_terminal;
+
+            // Check if terminal command contains {cmd} placeholder
+            if (strstr(terminal, "{cmd}") != NULL) {
+                // Replace all occurrences of {cmd} with the actual command
+                const char *placeholder = "{cmd}";
+                const size_t placeholder_len = 5; // strlen("{cmd}")
+                const size_t exec_len = strlen(a->exec);
+                const size_t terminal_len = strlen(terminal);
+
+                // Count occurrences of {cmd}
+                size_t occurrences = 0;
+                const char *pos = terminal;
+                while ((pos = strstr(pos, placeholder)) != NULL) {
+                    occurrences++;
+                    pos += placeholder_len;
+                }
+
+                // Calculate final string size
+                assert(terminal_len >= occurrences * placeholder_len); // should be true, so let's assert it is
+                size_t final_len = terminal_len - occurrences * placeholder_len + occurrences * exec_len + 1;
+                exec_with_terminal = xmalloc(final_len);
+
+                // Replace all {cmd} occurrences
+                const char *src = terminal;
+                char *dst = exec_with_terminal;
+                while ((pos = strstr(src, placeholder)) != NULL) {
+                    // Copy part before {cmd}
+                    size_t prefix_len = pos - src;
+                    memcpy(dst, src, prefix_len);
+                    dst += prefix_len;
+
+                    // Copy the exec command
+                    memcpy(dst, a->exec, exec_len);
+                    dst += exec_len;
+
+                    // Move past {cmd}
+                    src = pos + placeholder_len;
+                }
+
+                // Copy remaining part
+                strcpy(dst, src);
+            } else {
+                // Use the old behavior: "terminal command exec"
+                exec_with_terminal = xstrjoin3(terminal, " ", a->exec);
+            }
+
             free(a->exec);
             a->exec = exec_with_terminal;
         }
 
         char32_t *title = a->name;
         if (a != default_action) {
-            size_t title_len = c32len(default_action->name) +
-                3 +  /* “ — ” */
-                c32len(a->name) +
-                1;
-            title = malloc(title_len * sizeof(char32_t));
-
-            c32cpy(title, default_action->name);
-            c32cat(title, U" — ");
-            c32cat(title, a->name);
+            title = xc32join3(default_action->name, U" — ", a->name);
             free(a->name);
         }
 
-        tll_push_back(
-            *applications,
-            ((struct application){
-                .id = strdup(id),
-                .path = a->path,
-                .exec = a->exec,
-                .basename = c32dup(file_basename),
-                .wexec = a->wexec,
-                .title = title,
-                .app_id = a->app_id,
-                .generic_name = a->generic_name,
-                .comment = a->comment,
-                .keywords = a->keywords,
-                .categories = a->categories,
-                .icon = {.name = a->icon},
-                .visible = a->visible && (!filter_desktops || filter_desktop_entry(a, desktops)),
-                .count = 0}));
+        const size_t title_len = c32len(title);
+        const size_t basename_len = file_basename_lowercase != NULL
+            ? c32len(file_basename_lowercase)
+            : 0;
+        const size_t wexec_len = action->wexec != NULL
+            ? c32len(action->wexec)
+            : 0;
+        const size_t generic_name_len = action->generic_name != NULL
+            ? c32len(action->generic_name)
+            : 0;
+        const size_t comment_len = action->comment != NULL
+            ? c32len(action->comment)
+            : 0;
+
+        char32_t *title_lowercase = xc32dup(title);
+        for (size_t i = 0; i < title_len; i++)
+            title_lowercase[i] = toc32lower(title_lowercase[i]);
+
+        for (size_t i = 0; i < wexec_len; i++)
+            action->wexec[i] = toc32lower(action->wexec[i]);
+        for (size_t i = 0; i < generic_name_len; i++)
+            action->generic_name[i] = toc32lower(action->generic_name[i]);
+        for (size_t i = 0; i < comment_len; i++)
+            action->comment[i] = toc32lower(action->comment[i]);
+        tll_foreach(a->keywords, it) {
+            for (size_t i = 0; i < c32len(it->item); i++)
+                it->item[i] = toc32lower(it->item[i]);
+        }
+        tll_foreach(a->categories, it) {
+            for (size_t i = 0; i < c32len(it->item); i++)
+                it->item[i] = toc32lower(it->item[i]);
+        }
+
+        struct application *app = xmalloc(sizeof(*app));
+        *app = (struct application){
+            .id = xstrdup(id),
+            .index = 0,  /* Not used in application mode */
+            .path = a->path,
+            .exec = a->exec,
+            .app_id = a->app_id,
+            .title = title,
+            .title_lowercase = title_lowercase,
+            .basename = xc32dup(file_basename_lowercase),
+            .wexec = a->wexec,
+            .generic_name = a->generic_name,
+            .comment = a->comment,
+            .keywords = a->keywords,
+            .categories = a->categories,
+            .title_len = title_len,
+            .basename_len = basename_len,
+            .wexec_len = wexec_len,
+            .generic_name_len = generic_name_len,
+            .comment_len = comment_len,
+            .icon = {.name = a->icon},
+            .visible = a->visible && (!filter_desktops || filter_desktop_entry(a, desktops)),
+            .startup_notify = !a->no_startup_notify,
+            .count = 0,
+        };
+        tll_push_back(*applications, app);
+        tll_free_and_free(a->onlyshowin, free);
+        tll_free_and_free(a->notshowin, free);
     }
 
     free(id);
@@ -442,16 +532,9 @@ static char *
 new_id(const char *base_id, const char *new_part)
 {
     if (base_id == NULL)
-        return strdup(new_part);
+        return xstrdup(new_part);
 
-    size_t len = strlen(base_id) + 1 + strlen(new_part) + 1;
-    char *id = malloc(len);
-
-    strcpy(id, base_id);
-    strcat(id, "-");
-    strcat(id, new_part);
-
-    return id;
+    return xstrjoin3(base_id, "-", new_part);
 }
 
 static void
@@ -469,52 +552,35 @@ scan_dir(int base_fd, const char *terminal, bool include_actions,
     struct locale_variants lc_messages = {NULL};
 
     if (locale != NULL) {
-        char *copy = strdup(locale);
-
-        char *lang = copy;
-        char *country_start = strchr(copy, '_');
-        char *encoding_start = strchr(copy, '.');
-        char *modifier_start = strchr(copy, '@');
+        char *lang = xstrdup(locale);
+        char *country_start = strchr(lang, '_');
+        char *encoding_start = strchr(lang, '.');
+        char *modifier_start = strchr(lang, '@');
 
+        // Replace found delimiters with null terminators and skip past them
         if (country_start != NULL)
-            *country_start = '\0';
+            *country_start++ = '\0';
         if (encoding_start != NULL)
-            *encoding_start = '\0';
+            *encoding_start++ = '\0';
         if (modifier_start != NULL)
-            *modifier_start = '\0';
+            *modifier_start++ = '\0';
 
-        lc_messages.lang = copy;
-        if (country_start != NULL) {
-            if (asprintf(
-                    &lc_messages.lang_country, "%s_%s", lang, country_start + 1) < 0) {
-                LOG_WARN(
-                    "failed to construct lang_country string from %s and %s: %s",
-                    lang, country_start + 1, strerror(errno));
-            }
-        }
+        lc_messages.lang = lang;
+        if (country_start != NULL)
+            lc_messages.lang_country = xstrjoin3(lang, "_", country_start);
 
-        if (modifier_start != NULL) {
-            if (asprintf(
-                    &lc_messages.lang_modifier, "%s@%s", lang, modifier_start + 1) < 0) {
-                LOG_WARN(
-                    "failed to construct lang@modifier string from %s and %s: %s",
-                    lang, modifier_start + 1, strerror(errno));
-            }
-        }
+        if (modifier_start != NULL)
+            lc_messages.lang_modifier = xstrjoin3(lang, "@", modifier_start);
 
-        if (country_start != NULL && modifier_start != NULL) {
-            if (asprintf(
-                    &lc_messages.lang_country_modifier,
-                    "%s_%s@%s", lang, country_start + 1, modifier_start + 1) < 0) {
-                LOG_WARN(
-                    "failed to construct lang_country@modifier string from %s, %s and %s: %s",
-                    lang, country_start + 1, modifier_start + 1, strerror(errno));
-            }
-        }
+        if (country_start != NULL && modifier_start != NULL)
+            lc_messages.lang_country_modifier = xstrjoin3(
+                    lc_messages.lang_country, "@", modifier_start);
 
         LOG_DBG("lang=%s, lang_country=%s, lang@modifier=%s, lang_country@modifier=%s",
-                lc_messages.lang, lc_messages.lang_country,
-                lc_messages.lang_modifier, lc_messages.lang_country_modifier);
+                lc_messages.lang,
+                loggable_str(lc_messages.lang_country),
+                loggable_str(lc_messages.lang_modifier),
+                loggable_str(lc_messages.lang_country_modifier));
     }
 
 
@@ -524,12 +590,12 @@ scan_dir(int base_fd, const char *terminal, bool include_actions,
 
         struct stat st;
         if (fstatat(base_fd, e->d_name, &st, 0) == -1) {
-            LOG_WARN("%s: failed to stat", e->d_name);
+            LOG_WARN("%s: failed to stat: %s", e->d_name, strerror(errno));
             continue;
         }
 
         if (S_ISDIR(st.st_mode)) {
-            int dir_fd = openat(base_fd, e->d_name, O_RDONLY);
+            int dir_fd = openat(base_fd, e->d_name, O_RDONLY | O_CLOEXEC);
             if (dir_fd == -1) {
                 LOG_ERRNO("%s: failed to open", e->d_name);
                 continue;
@@ -552,9 +618,9 @@ scan_dir(int base_fd, const char *terminal, bool include_actions,
                 continue;
 
             //LOG_DBG("%s", e->d_name);
-            int fd = openat(base_fd, e->d_name, O_RDONLY);
+            int fd = openat(base_fd, e->d_name, O_RDONLY | O_CLOEXEC);
             if (fd == -1)
-                LOG_WARN("%s: failed to open", e->d_name);
+                LOG_WARN("%s: failed to open: %s", e->d_name, strerror(errno));
             else {
                 const char *file_basename = strrchr(e->d_name, '/');
                 if (file_basename == NULL)
@@ -562,7 +628,7 @@ scan_dir(int base_fd, const char *terminal, bool include_actions,
                 else
                     file_basename++;
 
-                const char *extension = strchr(file_basename, '.');
+                const char *extension = strrchr(file_basename, '.');
                 if (extension == NULL)
                     extension = file_basename + strlen(file_basename);
 
@@ -573,11 +639,14 @@ scan_dir(int base_fd, const char *terminal, bool include_actions,
                 mbsntoc32(wfile_basename, file_basename, extension - file_basename, chars);
                 wfile_basename[chars] = U'\0';
 
+                for (size_t i = 0; i < chars; i++)
+                    wfile_basename[i] = toc32lower(wfile_basename[i]);
+
                 char *id = new_id(base_id, e->d_name);
 
                 bool already_added = false;
                 tll_foreach(*applications, it) {
-                    if (strcmp(it->item.id, id) == 0) {
+                    if (strcmp(it->item->id, id) == 0) {
                         already_added = true;
                         break;
                     }
@@ -588,9 +657,11 @@ scan_dir(int base_fd, const char *terminal, bool include_actions,
                         fd, id, wfile_basename, terminal, include_actions,
                         filter_desktop, desktops,
                         &lc_messages, applications);
-                } else
+                    /* fd closed by parse_desktop_file() */
+                } else {
                     free(id);
-                close(fd);
+                    close(fd);
+                }
             }
         }
     }
@@ -605,9 +676,9 @@ scan_dir(int base_fd, const char *terminal, bool include_actions,
 static int
 sort_application_by_title(const void *_a, const void *_b)
 {
-    const struct application *a = _a;
-    const struct application *b = _b;
-    return c32cmp(a->title, b->title);
+    const struct application *const *a = _a;
+    const struct application *const *b = _b;
+    return c32casecmp((*a)->title, (*b)->title);
 }
 
 void
@@ -622,25 +693,34 @@ xdg_find_programs(const char *terminal, bool include_actions,
         char path[strlen(it->item.path) + 1 + strlen("applications") + 1];
         sprintf(path, "%s/applications", it->item.path);
 
-        int fd = open(path, O_RDONLY);
+        int fd = open(path, O_RDONLY | O_CLOEXEC);
         if (fd != -1) {
             scan_dir(fd, terminal, include_actions, filter_desktop, desktops, &apps, NULL);
             close(fd);
         }
     }
 
+    mtx_lock(&applications->lock);
     applications->count = tll_length(apps);
-    applications->v = malloc(tll_length(apps) * sizeof(applications->v[0]));
+    if (applications->count > 0) {
+        applications->v = xmalloc(tll_length(apps) * sizeof(applications->v[0]));
+    } else {
+        LOG_WARN("No applications found. See SEARCH PATHS in `man fuzzel` for details.");
+        applications->v = NULL;
+    }
 
     size_t i = 0;
     tll_foreach(apps, it) {
         applications->v[i++] = it->item;
+        if (it->item->visible)
+            applications->visible_count++;
         tll_remove(apps, it);
     }
     tll_free(apps);
 
     qsort(applications->v, applications->count, sizeof(applications->v[0]),
           &sort_application_by_title);
+    mtx_unlock(&applications->lock);
 
     xdg_data_dirs_destroy(dirs);
 
@@ -654,7 +734,7 @@ xdg_find_programs(const char *terminal, bool include_actions,
         int idx = 0;
         tll_foreach(app->keywords, it) {
             idx += swprintf(&keywords[idx],
-                            (sizeof(keywords) / sizeof(keywords[0])) - idx,
+                            ALEN(keywords) - idx,
                             L"%ls, ", (const wchar_t *)it->item);
         }
 
@@ -664,7 +744,7 @@ xdg_find_programs(const char *terminal, bool include_actions,
         idx = 0;
         tll_foreach(app->categories, it) {
             idx += swprintf(&categories[idx],
-                            (sizeof(categories) / sizeof(categories[0])) - idx,
+                            ALEN(categories) - idx,
                             L"%ls, ", (const wchar_t *)it->item);
         }
 
@@ -704,20 +784,16 @@ xdg_data_dirs(void)
     xdg_data_dirs_t ret = tll_init();
 
     const char *xdg_data_home = getenv("XDG_DATA_HOME");
+    const char *home;
     if (xdg_data_home != NULL && xdg_data_home[0] != '\0') {
-        int fd = open(xdg_data_home, O_RDONLY | O_DIRECTORY);
+        int fd = open(xdg_data_home, O_RDONLY | O_DIRECTORY | O_CLOEXEC);
         if (fd >= 0) {
-            struct xdg_data_dir d = {.fd = fd, .path = strdup(xdg_data_home)};
+            struct xdg_data_dir d = {.fd = fd, .path = xstrdup(xdg_data_home)};
             tll_push_back(ret, d);
         }
-    } else {
-        static const char *const local = ".local/share";
-        const struct passwd *pw = getpwuid(getuid());
-
-        char *path = malloc(strlen(pw->pw_dir) + 1 + strlen(local) + 1);
-        sprintf(path, "%s/%s", pw->pw_dir, local);
-
-        int fd = open(path, O_RDONLY | O_DIRECTORY);
+    } else if ((home = getenv("HOME")) != NULL && home[0] != '\0') {
+        char *path = xstrjoin(home, "/.local/share");
+        int fd = open(path, O_RDONLY | O_DIRECTORY | O_CLOEXEC);
         if (fd >= 0) {
             struct xdg_data_dir d = {.fd = fd, .path = path};
             tll_push_back(ret, d);
@@ -730,31 +806,31 @@ xdg_data_dirs(void)
     if (_xdg_data_dirs != NULL) {
 
         char *ctx = NULL;
-        char *copy = strdup(_xdg_data_dirs);
+        char *copy = xstrdup(_xdg_data_dirs);
 
         for (const char *tok = strtok_r(copy, ":", &ctx);
              tok != NULL;
              tok = strtok_r(NULL, ":", &ctx))
         {
-            int fd = open(tok, O_RDONLY | O_DIRECTORY);
+            int fd = open(tok, O_RDONLY | O_DIRECTORY | O_CLOEXEC);
             if (fd >= 0) {
-                struct xdg_data_dir d = {.fd = fd, .path = strdup(tok)};
+                struct xdg_data_dir d = {.fd = fd, .path = xstrdup(tok)};
                 tll_push_back(ret, d);
             }
         }
 
         free(copy);
     } else {
-        int fd1 = open("/usr/local/share", O_RDONLY | O_DIRECTORY);
-        int fd2 = open("/usr/share", O_RDONLY | O_DIRECTORY);
+        int fd1 = open("/usr/local/share", O_RDONLY | O_DIRECTORY | O_CLOEXEC);
+        int fd2 = open("/usr/share", O_RDONLY | O_DIRECTORY | O_CLOEXEC);
 
         if (fd1 >= 0) {
-            struct xdg_data_dir d = {.fd = fd1, .path = strdup("/usr/local/share")};
+            struct xdg_data_dir d = {.fd = fd1, .path = xstrdup("/usr/local/share")};
             tll_push_back(ret, d);
         }
 
         if (fd2 >= 0) {
-            struct xdg_data_dir d = {.fd = fd2, .path = strdup("/usr/share")};
+            struct xdg_data_dir d = {.fd = fd2, .path = xstrdup("/usr/share")};
             tll_push_back(ret, d);;
         }
     }
@@ -779,8 +855,12 @@ xdg_cache_dir(void)
     if (xdg_cache_home != NULL && xdg_cache_home[0] != '\0')
         return xdg_cache_home;
 
-    static char path[PATH_MAX];
-    const struct passwd *pw = getpwuid(getuid());
-    snprintf(path, sizeof(path), "%s/.cache", pw->pw_dir);
-    return path;
+    const char *home = getenv("HOME");
+    if (home != NULL && home[0] != '\0') {
+        static char path[PATH_MAX];
+        xsnprintf(path, sizeof(path), "%s/.cache", home);
+        return path;
+    }
+
+    return NULL;
 }
